import requests
import json

import pypowerbi.client
from pypowerbi.report import Report


class Reports:
    # url snippets
    groups_snippet = 'groups'
    reports_snippet = 'reports'
    rebind_snippet = 'rebind'
    clone_snippet = 'clone'
    generate_token_snippet = 'generatetoken'

    # json keys
    get_reports_value_key = 'value'

    def __init__(self, client):
        self.client = client
        self.base_url = f'{self.client.api_url}/{self.client.api_version_snippet}/{self.client.api_myorg_snippet}'

    def count(self, group_id=None):
        """
        Evaluates the number of reports
        :param group_id: The optional group id
        :return: The number of reports as returned by the API
        """
        return len(self.get_reports(group_id))

    def has_report(self, report_id, group_id=None):
        """
        Evaluates if the report exists
        :param report_id: The id of the report to evaluate
        :param group_id: The optional group id
        :return: True if the report exists, False otherwise
        """
        reports = self.get_reports(group_id)

        for report in reports:
            if report.id == report_id:
                return True

        return False

    def get_reports(self, group_id=None):
        """
        Gets all reports
        https://msdn.microsoft.com/en-us/library/mt634543.aspx
        :param group_id: The optional group id to get reports from
        :return: The list of reports for the given group
        """
        # group_id can be none, account for it
        if group_id is None:
            groups_part = '/'
        else:
            groups_part = f'/{self.groups_snippet}/{group_id}/'

        # form the url
        url = f'{self.base_url}{groups_part}{self.reports_snippet}/'
        # form the headers
        headers = self.client.auth_header

        # get the response
        response = requests.get(url, headers=headers)

        # 200 - OK. Indicates success. List of reports.
        if response.status_code == 200:
            reports = self.reports_from_get_reports_response(response)
        else:
            raise RuntimeError(f'Get reports request returned http error: {response.status_code}')

        return reports

    def get_report(self, report_id, group_id=None):
        """
        Gets a report
        https://msdn.microsoft.com/en-us/library/mt784668.aspx
        :param report_id: The id of the report to get
        :param group_id: The optional group id
        :return: The report as returned by the API
        """
        # # group_id can be none, account for it
        # if group_id is None:
        #     groups_part = '/'
        # else:
        #     groups_part = f'/{self.groups_snippet}/{group_id}/'
        #
        # # form the url
        # url = f'{self.base_url}{groups_part}{self.reports_snippet}/{report_id}'
        # # form the headers
        # headers = self.client.auth_header
        #
        # # get the response
        # response = requests.get(url, headers=headers)
        #
        # # 200 - OK. Indicates success.
        # if response.status_code != 200:
        #     raise RuntimeError(f'Get reports request returned http error: {response.status_code}')
        #
        # return Report.from_dict(json.loads(response.text))

        reports = self.get_reports(group_id)

        for report in reports:
            if report.id == report_id:
                return report

        raise RuntimeError('Could not find report')

    def clone_report(self, report_id, name, target_group_id, dataset_id, group_id=None):
        """
        Clones a report
        https://msdn.microsoft.com/en-us/library/mt784674.aspx
        :param report_id: The report id to clone
        :param name: The name to give the cloned report
        :param target_group_id: The target group for the cloned report
        :param dataset_id: The dataset id for the cloned report
        :param group_id: The optional group id
        :return: The cloned report
        """
        # group_id can be none, account for it
        if group_id is None:
            groups_part = '/'
        else:
            groups_part = f'/{self.groups_snippet}/{group_id}/'

        # form the url
        url = f'{self.base_url}{groups_part}{self.reports_snippet}/{report_id}/{self.clone_snippet}'
        # form the headers
        headers = self.client.auth_header
        # form the json
        json_dict = {
            Report.name_key: name,
            Report.target_model_id_key: str(dataset_id),
        }

        # target group id can be none, account for it
        if target_group_id is not None:
            json_dict[Report.target_workspace_id_key] = str(target_group_id)

        # get the response
        response = requests.post(url, headers=headers, json=json_dict)

        # 200 - OK. Indicates success.
        if response.status_code != 200:
            raise RuntimeError(f'Clone reports request returned http error: {response.status_code}')

        return Report.from_dict(json.loads(response.text))

    def delete_report(self, report_id, group_id=None):
        """
        Deletes a report
        https://msdn.microsoft.com/en-us/library/mt784671.aspx
        :param report_id: The id of the report to delete
        :param group_id: The id of the group from which to delete the report
        """
        # group_id can be none, account for it
        if group_id is None:
            groups_part = '/'
        else:
            groups_part = f'/{self.groups_snippet}/{group_id}/'

        # form the url
        url = f'{self.base_url}{groups_part}{self.reports_snippet}/{report_id}/'
        # form the headers
        headers = self.client.auth_header

        # get the response
        response = requests.delete(url, headers=headers)

        # 200 - OK. Indicates success.
        if response.status_code != 200:
            pass

    def rebind_report(self, report_id, dataset_id, group_id=None):
        """
        Rebinds a report to another dataset
        https://msdn.microsoft.com/en-us/library/mt784672.aspx
        :param report_id: The id of the report to rebind
        :param dataset_id: The id of the dataset to rebind the report to
        :param group_id: The optional id of the group from which the report belongs to
        """
        # group_id can be none, account for it
        if group_id is None:
            groups_part = '/'
        else:
            groups_part = f'/{self.groups_snippet}/{group_id}/'

        # form the url
        url = f'{self.base_url}{groups_part}{self.reports_snippet}/{report_id}/{self.rebind_snippet}'
        # form the headers
        headers = self.client.auth_header
        # form the json
        json_dict = {
            Report.dataset_id_key: dataset_id
        }

        # get the response
        response = requests.post(url, headers=headers, json=json_dict)

        # 200 - OK. Indicates success.
        if response.status_code != 200:
            raise RuntimeError(f'Rebind report request returned http error: {response.status_code}')

    def generate_token(self, report_id, token_request, group_id):
        """
        Generates an embed token for a report
        https://msdn.microsoft.com/en-us/library/mt784614.aspx
        :param report_id: The report to generate teh token for
        :param token_request: The token request object
        :param group_id: The group id
        :return: Returns the embed token
        """
        # form the url
        url = f'{self.base_url}/{self.groups_snippet}/{group_id}/' \
              f'{self.reports_snippet}/{report_id}/{self.generate_token_snippet}'
        # form the headers
        headers = self.client.auth_header
        # form the json
        json_dict = pypowerbi.client.TokenRequestEncoder().default(token_request)

        # get the response
        response = requests.post(url, headers=headers, json=json_dict)

        # 200 - OK. Indicates success.
        if response.status_code != 200:
            raise RuntimeError(f'Generate token for report request returned http error: {response.status_code}')

        return pypowerbi.client.EmbedToken.from_dict(json.loads(response.text))

    @classmethod
    def reports_from_get_reports_response(cls, response):
        """
        Creates a list of reports from a http response
        :param response: The response to create the reports from
        :return: A list of reports created from the http response
        """
        # load the response into a dict
        response_dict = json.loads(response.text)
        reports = []
        # go through entries returned from API
        for entry in response_dict[cls.get_reports_value_key]:
            reports.append(Report.from_dict(entry))

        return reports



import json


class Dataset:
    # json keys
    id_key = 'id'
    name_key = 'name'
    add_rows_api_enabled_key = 'addRowsAPIEnabled'
    configured_by_key = 'configuredBy'
    is_refreshable_key = 'isRefreshable'
    is_effective_identity_required_key = 'isEffectiveIdentityRequired'
    is_effective_identity_roles_required_key = 'isEffectiveIdentityRolesRequired'
    is_on_prem_gateway_required_key = 'isOnPremGatewayRequired'
    tables_key = 'tables'

    def __init__(self, name, dataset_id=None, tables=None, add_rows_api_enabled=None,
                 configured_by=None, is_refreshable=None, is_effective_identity_required=None,
                 is_effective_identity_roles_required=None, is_on_prem_gateway_required=None):
        self.name = name
        self.id = dataset_id
        self.tables = tables
        self.add_rows_api_enabled = add_rows_api_enabled
        self.configured_by = configured_by
        self.is_refreshable = is_refreshable
        self.is_effective_identity_required = is_effective_identity_required
        self.is_effective_identity_roles_required = is_effective_identity_roles_required
        self.is_on_prem_gateway_required = is_on_prem_gateway_required

    @classmethod
    def from_dict(cls, dictionary):
        """
        Creates a dataset from a dictionary, key values for 'id' and 'name' required
        :param dictionary: The dictionary to create the dataset from
        :return: A dataset created from the given dictionary
        """
        # id is required
        if Dataset.id_key in dictionary:
            dataset_id = str(dictionary[Dataset.id_key])
            # id cannot be whitespace
            if dataset_id.isspace():
                raise RuntimeError('Dataset dict has empty id key value')
        else:
            raise RuntimeError('Dataset dict has no id key')
        # name is required
        if Dataset.name_key in dictionary:
            dataset_name = str(dictionary[Dataset.name_key])
            # name cannot be whitespace
            if dataset_id.isspace():
                raise RuntimeError('Dataset dict has empty name key value')
        else:
            raise RuntimeError('Dataset dict has no name key')

        # add api enabled is optional
        if Dataset.add_rows_api_enabled_key in dictionary:
            add_rows_api_enabled = bool(dictionary[Dataset.add_rows_api_enabled_key])
        else:
            add_rows_api_enabled = None

        # configured by is optional
        if Dataset.configured_by_key in dictionary:
            configured_by = str(dictionary[Dataset.configured_by_key])
        else:
            configured_by = None

        # is refreshable is optional
        if Dataset.is_refreshable_key in dictionary:
            is_refreshable = bool(dictionary[Dataset.is_refreshable_key])
        else:
            is_refreshable = None

        # is effective identity required is optional
        if Dataset.is_effective_identity_required_key in dictionary:
            is_effective_identity_required = bool(dictionary[Dataset.is_effective_identity_required_key])
        else:
            is_effective_identity_required = None

        # is effective identity roles required is optional
        if Dataset.is_effective_identity_roles_required_key in dictionary:
            is_effective_identity_roles_required = bool(dictionary[Dataset.is_effective_identity_roles_required_key])
        else:
            is_effective_identity_roles_required = None

        # is on prem gateway required is optional
        if Dataset.is_on_prem_gateway_required_key in dictionary:
            is_on_prem_gateway_required = bool(dictionary[Dataset.is_on_prem_gateway_required_key])
        else:
            is_on_prem_gateway_required = None

        return Dataset(dataset_name, dataset_id, add_rows_api_enabled=add_rows_api_enabled,
                       configured_by=configured_by, is_refreshable=is_refreshable,
                       is_effective_identity_required=is_effective_identity_required,
                       is_effective_identity_roles_required=is_effective_identity_roles_required,
                       is_on_prem_gateway_required=is_on_prem_gateway_required)


class DatasetEncoder(json.JSONEncoder):
    def default(self, o):
        table_encoder = TableEncoder()

        json_dict = {
            Dataset.name_key: o.name,
            Dataset.tables_key: [table_encoder.default(x) for x in o.tables],
        }

        return json_dict


class Table:
    name_key = 'name'
    columns_key = 'columns'
    measures_key = 'measures'

    @classmethod
    def from_dict(cls, dictionary):
        """
        Creates a table from a dictionary, 'name' key value required
        :param dictionary: The dictionary to create the table from
        :return: A table created from the dictionary
        """
        # name is required
        if Table.name_key in dictionary:
            table_name = str(dictionary[Table.name_key])
            # name cannot be whitespace
            if table_name.isspace():
                raise RuntimeError('Table dict has empty name key value')
        else:
            raise RuntimeError('Table dict has no name key')

        # measures are optional
        if Table.measures_key in dictionary:
            table_measures = [Table.from_dict(x) for x in dictionary[Table.measures_key]]
        else:
            table_measures = None

        return Table(name=table_name, measures=table_measures)

    def __init__(self, name, columns=None, measures=None):
        self.name = name
        self.columns = columns
        self.measures = measures


class TableEncoder(json.JSONEncoder):
    def default(self, o):
        json_dict = {
            Table.name_key: o.name,
        }

        if o.columns is not None:
            column_encoder = ColumnEncoder()
            json_dict[Table.columns_key] = [column_encoder.default(x) for x in o.columns]

        if o.measures is not None:
            measure_encoder = MeasureEncoder()
            json_dict[Table.measures_key] = [measure_encoder.default(x) for x in o.measures]

        return json_dict


class Measure:
    name_key = 'name'
    expression_key = 'expression'
    formatstring_key = 'formatString'
    is_hidden_key = 'isHidden'

    @classmethod
    def from_dict(cls, dictionary):
        # name is required
        if Measure.name_key in dictionary:
            measure_name = str(dictionary[Measure.name_key])
            # name cannot be whitespace
            if measure_name.isspace():
                raise RuntimeError('Measure dict has empty name key value')
        else:
            raise RuntimeError('Measure dict has no name key')

        # expression is required
        if Measure.expression_key in dictionary:
            measure_expression = str(dictionary[Measure.expression_key])
            # expression cannot be whitespace
            if measure_expression.isspace():
                raise RuntimeError('Measure dict has empty expression key value')
        else:
            raise RuntimeError('Measure dict has no expression key')

        if Measure.formatstring_key in dictionary:
            measure_formatstring = str(dictionary[Measure.formatstring_key])
        else:
            measure_formatstring = None

        if Measure.is_hidden_key in dictionary:
            measure_is_hidden = bool(dictionary[Measure.is_hidden_key])
        else:
            measure_is_hidden = None

        return Measure(name=measure_name, expression=measure_expression, formatstring=measure_formatstring,
                          is_hidden=measure_is_hidden)

    def __init__(self, name, expression, formatstring=None, is_hidden=None):
        self.name = name
        self.expression = expression
        self.formatstring = formatstring
        self.is_hidden = is_hidden


class MeasureEncoder(json.JSONEncoder):
    def default(self, o):
        json_dict = {
            Measure.name_key: o.name,
            Measure.expression_key: o.expression,
        }

        if o.formatstring is not None:
            json_dict[Measure.formatstring_key] = o.formatstring

        if o.is_hidden is not None:
            json_dict[Measure.is_hidden_key] = o.is_hidden

        return json_dict


class Column:
    name_key = 'name'
    datatype_key = 'dataType'

    def __init__(self, name, data_type):
        self.name = name
        self.data_type = data_type


class ColumnEncoder(json.JSONEncoder):
    def default(self, o):
        return {
            Column.name_key: o.name,
            Column.datatype_key: o.data_type
        }


class Row:
    def __init__(self, **kwargs):
        for key in kwargs:
            setattr(self, key, kwargs[key])


class RowEncoder(json.JSONEncoder):
    def default(self, o):
        return o.__dict__



import json


class Report:
    id_key = 'id'
    name_key = 'name'
    web_url_key = 'webUrl'
    embed_url_key = 'embedUrl'
    dataset_id_key = 'datasetId'
    target_workspace_id_key = 'targetWorkspaceId'
    target_model_id_key = 'targetModelId'

    def __init__(self, report_id, name, web_url, embed_url, dataset_id):
        self.id = report_id
        self.name = name
        self.web_url = web_url
        self.embed_url = embed_url
        self.dataset_id = dataset_id

    @classmethod
    def from_dict(cls, dictionary):
        """
        Creates a report from a dictionary
        :param dictionary: The dictionary to create a report from
        :return: The created dictionary
        """
        # id is required
        if cls.id_key in dictionary:
            report_id = str(dictionary[cls.id_key])
            # id cannot be whitespace
            if report_id.isspace():
                raise RuntimeError(f'Report dict has empty {cls.id_key} key value')
        else:
            raise RuntimeError(f'Report dict has no {cls.id_key} key')

        # name is required
        if cls.name_key in dictionary:
            report_name = str(dictionary[cls.name_key])
            # name cannot be whitespace
            if report_name.isspace():
                raise RuntimeError(f'Report dict has empty {cls.name_key} key value')
        else:
            raise RuntimeError(f'Report dict has no {cls.name_key} key')

        # web url is optional
        if cls.web_url_key in dictionary:
            web_url = str(dictionary[cls.web_url_key])
        else:
            web_url = None

        # embed url is optional
        if cls.embed_url_key in dictionary:
            embed_url = str(dictionary[cls.embed_url_key])
        else:
            embed_url = None

        # dataset id is required
        if cls.dataset_id_key in dictionary:
            dataset_id = str(dictionary[cls.dataset_id_key])
            # dataset id cannot be whitespace
            if dataset_id.isspace():
                raise RuntimeError(f'Report dict has empty {cls.dataset_id_key} key value')
        else:
            raise RuntimeError(f'Report dict has no {cls.dataset_id_key} key')

        return Report(report_id, report_name, web_url, embed_url, dataset_id)


class ReportEncoder(json.JSONEncoder):
    def default(self, o):
        return {
            Report.id_key: o.id,
            Report.name_key: o.name,
            Report.web_url_key: o.web_url,
            Report.embed_url_key: o.embed_url,
            Report.dataset_id_key: o.dataset_id
        }



from setuptools import setup

setup(name='pypowerbi',
      version='0.15',
      description='A python library for Microsoft\'s PowerBI',
      url='http://github.com/cmberryau/pypowerbi',
      author='Chris Berry',
      author_email='chris@chrisberry.com.au',
      license='MIT',
      packages=['pypowerbi'],
      install_requires=[
            'requests',
            'adal',
      ],
      zip_safe=False)



import json
import datetime

from .reports import Reports
from .datasets import Datasets


class PowerBIClient:
    default_resource_url = 'https://analysis.windows.net/powerbi/api'
    default_api_url = 'https://api.powerbi.com'

    api_version_snippet = 'v1.0'
    api_myorg_snippet = 'myorg'

    def __init__(self, api_url, token):
        self.api_url = api_url
        self.token = token
        self.datasets = Datasets(self)
        self.reports = Reports(self)

    @property
    def auth_header(self):
        if self._auth_header is None:
            self._auth_header = {
                'Authorization': f'Bearer {self.token["accessToken"]}'
            }

        return self._auth_header

    _auth_header = None


class EffectiveIdentity:
    username_key = 'username'
    roles_key = 'roles'
    datasets_key = 'datasets'

    def __init__(self, username, roles, datasets):
        self.username = username
        self.roles = roles
        self.datasets = datasets


class EffectiveIdentityEncoder(json.JSONEncoder):
    def default(self, o):
        return {
            EffectiveIdentity.username_key: o.username,
            EffectiveIdentity.roles_key: o.roles,
            EffectiveIdentity.datasets_key: o.datasets,
        }


class TokenRequest:
    access_level_key = 'accessLevel'
    dataset_id_key = 'datasetId'
    allow_saveas_key = 'allowSaveAs'
    identities_key = 'identities'

    def __init__(self, access_level, dataset_id=None, allow_saveas=None, identities=None):
        self.access_level = access_level
        self.dataset_id = dataset_id
        self.allow_saveas = allow_saveas
        self.identities = identities


class TokenRequestEncoder(json.JSONEncoder):
    def default(self, o):
        effective_identity_encoder = EffectiveIdentityEncoder()

        json_dict = {
            TokenRequest.access_level_key: o.access_level
        }

        if o.dataset_id is not None:
            json_dict[TokenRequest.dataset_id_key] = o.dataset_id

        if o.allow_saveas is not None:
            json_dict[TokenRequest.allow_saveas_key] = o.allow_saveas

        if o.identities is not None:
            json_dict[TokenRequest.identities_key] = [effective_identity_encoder.default(x) for x in o.identities]

        return json_dict


class EmbedToken:
    token_key = 'token'
    token_id_key = 'tokenId'
    expiration_key = 'expiration'

    def __init__(self, token, token_id, expiration):
        self.token = token
        self.token_id = token_id
        self.expiration = expiration

    @classmethod
    def from_dict(cls, dictionary):
        if cls.token_key not in dictionary:
            raise RuntimeError(f'Token dict has no {cls.token_key} key')

        token = dictionary[cls.token_key]
        token_id = dictionary[cls.token_id_key]
        expiration = dictionary[cls.expiration_key]

        return EmbedToken(token, token_id, expiration)

    @property
    def expiration_as_datetime(self):
        return datetime.datetime.strptime(self.expiration, '%Y-%m-%dT%H:%M:%SZ')
